import pandas as pd
import os
from unidecode import unidecode
from openpyxl import load_workbook


def QuitarPuntosDecimalesTextos(df, columnas):
    """Convierte las columnas especificadas a string y elimina el sufijo '.0' de los valores numéricos.
    Returns:
        pd.DataFrame: El DataFrame con las columnas modificadas.
    """
    df[columnas] = df[columnas].astype(str).replace(r'^(\d+)\.0$', r'\1', regex=True)
    return df


def mergeBases(dataframe1, dataframe2, lefton, righton):
    """Une dos DataFrames de Pandas asegurando que no haya duplicados en el cruce y detectando registros sin coincidencias en cada DataFrame.

    Parámetros:
    - dataframe1: Primer DataFrame de Pandas.
    - dataframe2: Segundo DataFrame de Pandas.
    - lefton: Lista de columnas clave en dataframe1.
    - righton: Lista de columnas clave en dataframe2.

    Retorna:
    - dfmerge: DataFrame con la unión.
    - df_no_match_left: Registros en dataframe1 que no encontraron match en dataframe2.
    - df_no_match_right: Registros en dataframe2 que no encontraron match en dataframe1.
    - alerta_duplicados: Mensaje de alerta si hay duplicados en la clave.
    """

    # Realizar el merge con how='left'
    if lefton == righton:
        dfmerge = pd.merge(dataframe1, dataframe2, on=lefton, how='left',suffixes=('','_right'))
    else:
        dfmerge = pd.merge(dataframe1, dataframe2, left_on=lefton, right_on=righton, how='left', suffixes=('','_right'))

    # Eliminar columnas duplicadas generadas por el merge
    if isinstance(righton, list):  # Si righton es una lista de columnas
        for col in righton:
            columna_duplicada = f"{col}_right"
            if columna_duplicada in dfmerge.columns:
                dfmerge = dfmerge.drop(columns=[columna_duplicada])
    else:  # Si righton es una sola columna
        columna_duplicada = f"{righton}_right"
        if columna_duplicada in dfmerge.columns:
            dfmerge = dfmerge.drop(columns=[columna_duplicada])

    # Detectar registros de df1 que no tienen coincidencias en df2 (valores NaN en columnas de df2)
    columnas_df2 = [col for col in dataframe2.columns if col not in righton]
    df_no_match_left = dfmerge[dfmerge[columnas_df2].isna().all(axis=1)]

    # Detectar registros de df2 que no tienen coincidencias en df1
    claves_merge = set(dfmerge[lefton].drop_duplicates().itertuples(index=False, name=None))
    df_no_match_right = dataframe2[~dataframe2[righton].apply(lambda x: tuple(x) in claves_merge, axis=1)]

    # Detectar duplicados en las claves del merge
    duplicados = dfmerge.groupby(lefton).size()
    duplicados = duplicados[duplicados > 1]  # Filtrar claves duplicadas

    alerta_duplicados = "⚠️ Hay duplicados en la clave del cruce" if len(duplicados) > 0 else "✅ No hay duplicados en la clave"

    return dfmerge, df_no_match_left, df_no_match_right, alerta_duplicados


# Quita las tildes de un DataFrame
def dfSinTildes(dataframe):
    for columna in dataframe.columns:
        if dataframe[columna].dtype == 'object':
            dataframe[columna] = dataframe[columna].apply(unidecode)
    return  dataframe

# Quita las tildes de los encabezados de un DataFrame        
def dfEncabezadosSinTildes(dataframe):
    nuevos_nombres_columnas = [unidecode(columna) 
        if dataframe[columna].dtype == 'object' else columna for columna in dataframe.columns]
    dataframe.columns = nuevos_nombres_columnas
    return  dataframe

def GuardarRemplazarExcel(dataframe,rutaParaGuardado,nombreHoja, ejecutar=True):  
    if ejecutar:
        if os.path.exists(rutaParaGuardado):
            os.remove(rutaParaGuardado)
        dataframe.to_excel(rutaParaGuardado,sheet_name=nombreHoja, index=False)

def GuardarRemplazarCsv(dataframe,rutaParaGuardado, ejecutar=True):  
    if ejecutar:
        if os.path.exists(rutaParaGuardado):
            os.remove(rutaParaGuardado)
        dataframe.to_csv(rutaParaGuardado, index=False)

def GuardarRemplazarParquet(dataframe,rutaParaGuardado, ejecutar=True, engine="fastparquet"):  
    if ejecutar:
        if os.path.exists(rutaParaGuardado):
            os.remove(rutaParaGuardado)
        if engine=="":    
            dataframe.to_parquet(rutaParaGuardado, index=False)
        else:
            dataframe.to_parquet(rutaParaGuardado, index=False, engine=engine)


def EstandarizarColumnasConDatosBool(dataframe):
    for col in dataframe.columns:
        # Si en la columna hay al menos un valor booleano
        if dataframe[col].apply(lambda x: isinstance(x, bool)).any():
            # Reemplazar NaN con False y convertir todo a tipo bool
            dataframe[col] = dataframe[col].astype(object).fillna(False).astype(bool)
    return dataframe

## Funcion que se encarga de separar el nombre del archivo con '-', al comienzo del archivo se almacena el mes de la informacion, por lo que con ello se tiene el mes para cada uno de los dataframes
def extraerMesInicioDelNombre(nombreArchivo,Separador):
    """ Extrae el número y nombre del mes desde el nombre del archivo. """
    nombreArchivo = nombreArchivo.replace(" ", "")  # Elimina espacios extra
    partes = nombreArchivo.split(Separador)
    if len(partes) >= 1:
        numero_mes = partes[0].strip()  # "01"
        return numero_mes

    return None, None  # Retornar valores nulos si el formato no es el esperado

## Funcion que convierte una cadena a (UpperCamelCase), eliminando espacios y guiones bajos.
def pasarACamelCase(texto,Separador):
    # Reemplaza guiones bajos por espacios, divide en palabras, y las capitaliza
    partes = texto.replace(Separador, " ").split()
    return ''.join(p.capitalize() for p in partes)

def obtenerHojasVisiblesExcel(RutaArchivo):
    """Obtiene solo las hojas visibles de un archivo Excel (.xlsx)."""
    try:
        wb = load_workbook(RutaArchivo, read_only=True)
        hojas_visibles = [sheet.title for sheet in wb.worksheets if sheet.sheet_state == "visible"]
        wb.close()
        return hojas_visibles
    except Exception as e:
        print(f"Error al leer hojas de {RutaArchivo}: {e}")
        return []

def obtenerTodasLasHojasExcel(RutaArchivo):
    """Obtiene todas las hojas de un archivo Excel (.xlsx), sin importar si están visibles u ocultas."""
    try:
        wb = load_workbook(RutaArchivo, read_only=True)
        hojas = [sheet.title for sheet in wb.worksheets]  # Todas las hojas
        wb.close()
        return hojas
    except Exception as e:
        print(f"Error al leer hojas de {RutaArchivo}: {e}")
        return []

    ##  Renombra las columnas de un DataFrame aplicando formato PascalCase al nombre original,y agrega el nombre original al final separado por un guion bajo.
def renombraColumnasColumnasTablasConNombreHoja(dataframe,Separador):
    nuevas_columnas = {}
    for col in dataframe.columns:
        if "_" in col:
            izquierda, derecha = col.split("_", 1)  # Solo divide en el primer guion bajo
            nuevo_nombre = f"{pasarACamelCase(izquierda,Separador)}_{derecha}"
        else:
            nuevo_nombre = pasarACamelCase(col,Separador)
        nuevas_columnas[col] = nuevo_nombre
    return dataframe.rename(columns=nuevas_columnas)


import numpy as np
import pandas as pd
import os 
import shutil
from unidecode import unidecode
import openpyxl
import time
import pythoncom
import win32com.client

# Borrar los archivos si existen en la carpeta de bases Banrep
def BorrrarArchivo(rutaArchivo):  
    if os.path.exists(rutaArchivo):# Verificar si el archivo ya existe
        os.remove(rutaArchivo)

# Mueve los archivos descargados a la carpeta que se requiere
def moverArchivo(RutaOrigen, RutaDestino):
    try:
        shutil.move(RutaOrigen, RutaDestino)
        print(f"El archivo '{RutaOrigen}' se ha movido correctamente a '{RutaDestino}'.")
    except Exception as e:
        print(f"No se pudo mover el archivo '{RutaOrigen}' a '{RutaDestino}': {e}")

def copiarArchivo(RutaOrigen, RutaDestino):
    try:
        shutil.copy2(RutaOrigen, RutaDestino)  # `copy2` copia metadatos (marca de tiempo) además del archivo
        print(f"El archivo '{RutaOrigen}' se ha copiado correctamente a '{RutaDestino}'.")
    except Exception as e:
        print(f"No se pudo copiar el archivo '{RutaOrigen}' a '{RutaDestino}': {e}")


def obtenerArchivoMasReciente(rutaCarpeta, formato=".txt"):
    # Obtener la lista de archivos en la carpeta
    archivos_en_carpeta = [archivo for archivo in os.listdir(rutaCarpeta) if archivo.endswith(formato)]
    # Verificar si hay archivos en la carpeta
    if archivos_en_carpeta:
        # Obtener el archivo más reciente basado en la última modificación
        archivo_mas_reciente = max(archivos_en_carpeta,key=lambda x: os.path.getmtime(os.path.join(rutaCarpeta, x)))
        # Obtener la fecha de modificación del archivo más reciente
        fecha_modificacion = os.path.getmtime(os.path.join(rutaCarpeta, archivo_mas_reciente))
        # Retornar nombre y fecha
        return archivo_mas_reciente, fecha_modificacion
    else:
        return None, None

def ExisteRuta(arregloDeCarpetas):
    carpeta_correcta = None
    for carpeta in arregloDeCarpetas:
        try:
            # Verifica si la ruta existe y es una carpeta
            if os.path.isdir(carpeta):
                print(f"La carpeta existe en: {carpeta}")
                carpeta_correcta = carpeta
                break  # Sale del ciclo si la carpeta existe
        except Exception as e:
            # Maneja otras posibles excepciones (e.g., problemas de permisos)
            print(f"Error al verificar la carpeta {carpeta}: {e}")
            continue
    if carpeta_correcta is None:
        print("La carpeta no se encontró en ninguna de las rutas proporcionadas.")
    else:
        # Aquí puedes continuar con el procesamiento usando la carpeta_correcta
        pass
    return carpeta_correcta

def cargaExcel(rutaArchivo,sheetName, AsString=False):
    df = pd.read_excel(rutaArchivo, sheet_name=sheetName, engine='openpyxl')
    if AsString:
        df=df.astype(str)
    return df

def ActualizarTodoExcel(ruta_archivo):
    pythoncom.CoInitialize()  # Inicializa COM correctamente

    try:
        ruta_archivo_absoluta = os.path.abspath(ruta_archivo)
        print(f"Intentando abrir el archivo: {ruta_archivo_absoluta}")

        if not os.path.exists(ruta_archivo_absoluta):
            print(f"⚠️ El archivo no existe en la ruta: {ruta_archivo_absoluta}")
            return
        # Iniciar Excel y hacerlo visible
        excel_app = win32com.client.DispatchEx("Excel.Application")
        excel_app.Visible = True  
        excel_app.DisplayAlerts = False
        # Abrir el archivo de Excel
        workbook = excel_app.Workbooks.Open(ruta_archivo_absoluta)
        print("📂 Archivo abierto correctamente, iniciando actualización...")

        # Activar la primera hoja (por si acaso)
        workbook.Sheets(1).Activate()
        # Iniciar actualización
        workbook.RefreshAll()
        print("🔄 Actualizando conexiones y tablas de Power Query...")
        # Esperar que Excel termine la actualización y cálculos
        while excel_app.CalculationState != 0:  # 0 significa "listo"
            print("⌛ Esperando que Excel termine los cálculos...")
            time.sleep(2)
        print("✅ Actualización completada.")
        # Guardar y cerrar Excel
        print(f'📁 Archivo actualizado y cerrado: {ruta_archivo_absoluta}')
    except Exception as e:
        print(f"❌ Error al actualizar el archivo {ruta_archivo_absoluta}: {e}")
    finally:
        # Asegurar que Excel se cierra correctamente
        if 'excel_app' in locals():
            a=10
        pythoncom.CoUninitialize()

def consolidarArchivosExcel(NombreConElQueIniciaLosArchivos,RutaALaCarpeta,TipoArchivos='.xlsx'):
    dataframes = []
    for archivo in os.listdir(RutaALaCarpeta):
        if archivo.endswith(TipoArchivos) and archivo.startswith(NombreConElQueIniciaLosArchivos):
            ruta_archivo = os.path.join(RutaALaCarpeta, archivo)
            df = pd.read_excel(ruta_archivo)
            dataframes.append(df)    
    df_consolidado = pd.concat(dataframes, ignore_index=True)
    return df_consolidado




import numpy as np

def corregirSiAplicaExcp(df,ResultadoSinExp,ColumnaExcepcion,ColumnaAfectaRanking):
    df[ColumnaExcepcion] = np.where((df[ResultadoSinExp]==True),False,df[ColumnaExcepcion])
    df[ColumnaAfectaRanking] = np.where((df[ResultadoSinExp]==True),False,df[ColumnaAfectaRanking])
